@@ -4,15 +4,17 @@ import base64
 import datetime
 import logging
 import threading
-from logging import LogRecord
+from enum import Enum
+from logging import ERROR, LogRecord
 from typing import Any, Dict, List, Optional, cast
+from uuid import uuid4
 
 import attr
 import curlify
 import requests
 
 from .events import Payload
-from .failures import FailureContext
+from .failures import Evidence
 from .status import Status
 from .types import Headers
 from .utils import format_exception
@@ -25,19 +27,46 @@ class Case:
     curl_code: Optional[str] = attr.ib(default=None)
 
 
+class Risk(Enum):
+    """Risk on a failed assertion."""
+
+    low = "Low"
+    medium = "Medium"
+    high = "High"
+
+    def __str__(self) -> str:
+        return self.value
+
+
+class Confidence(Enum):
+    """Confidence of an assertion."""
+
+    low = "Low"
+    medium = "Medium"
+    high = "High"
+
+    def __str__(self) -> str:
+        return self.value
+
+
 @attr.s(slots=True, repr=False)
-class Check:
-    """Single check run result."""
+class AssertionResult:
+    """Assertion result."""
 
     name: str = attr.ib()
-    value: Status = attr.ib()
-    response: Optional[requests.Response] = attr.ib()
-    elapsed: float = attr.ib()
-    example: Optional[Case] = attr.ib(default=None)
+    status: Status = attr.ib()
+    interactions: List[Interaction] = attr.ib(factory=list)
+    elapsed: float = attr.ib(factory=float)
+    confidence: Confidence = attr.ib(default=Confidence.low)
+    risk: Risk = attr.ib(default=Risk.low)
+    id: str = attr.ib(factory=(lambda: str(uuid4())))
+    recorded_at: str = attr.ib(factory=lambda: datetime.datetime.now().isoformat())
+    evidence: Optional[Evidence] = attr.ib(default=None)
+    reference: Optional[str] = attr.ib(default=None)
+    solution: Optional[str] = attr.ib(default=None)
     message: Optional[str] = attr.ib(default=None)
-    # Failure-specific context
-    context: Optional[FailureContext] = attr.ib(default=None)
-    request: Optional[requests.PreparedRequest] = attr.ib(default=None)
+    code: Optional[str] = attr.ib(default=None)
+    cwe: Optional[str] = attr.ib(factory=str)
 
 
 @attr.s(slots=True, repr=False)
@@ -74,7 +103,10 @@ class Request:
         prepared_request.prepare(
             url=self.uri,
             method=self.method,
-            headers=self.headers,
+            headers={
+                key: ";".join(value) if isinstance(value, list) else value
+                for (key, value) in self.headers.items()
+            },
             data=base64.b64decode(self.body.encode()).decode()
             if self.body is not None
             else self.body,
@@ -94,10 +126,11 @@ class Response:
     status_code: int = attr.ib()
     message: str = attr.ib()
     headers: Dict[str, List[str]] = attr.ib()
+    method: str = attr.ib()
+    uri: str = attr.ib()
     body: Optional[str] = attr.ib()
     encoding: Optional[str] = attr.ib()
     http_version: str = attr.ib()
-    elapsed: float = attr.ib()
 
     @classmethod
     def from_requests(cls, response: requests.Response) -> "Response":
@@ -121,11 +154,12 @@ class Response:
         return cls(
             status_code=response.status_code,
             message=response.reason,
+            method=response.request.method if response.request.method else "",
+            uri=cast(str, response.request.url),
             body=body,
             encoding=response.encoding,
             headers=headers,
             http_version=http_version,
-            elapsed=response.elapsed.total_seconds(),
         )
 
 
@@ -135,31 +169,101 @@ class Interaction:
 
     request: Request = attr.ib()
     response: Response = attr.ib()
-    checks: List[Check] = attr.ib()
     status: Status = attr.ib()
+    elapsed: float = attr.ib()
+    id: str = attr.ib(factory=(lambda: str(uuid4())))
+    name: Optional[str] = attr.ib(factory=str)
     recorded_at: str = attr.ib(factory=lambda: datetime.datetime.now().isoformat())
 
     @classmethod
     def from_requests(
-        cls, response: requests.Response, status: Status, checks: List[Check]
+        cls, response: requests.Response, status: Status
     ) -> "Interaction":
         return cls(
             request=Request.from_prepared_request(response.request),
             response=Response.from_requests(response),
             status=status,
-            checks=checks,
+            elapsed=response.elapsed.total_seconds() * 1000,
         )
 
 
+class StepRecordType(Enum):
+    assertion = "assertion"
+    interaction = "interaction"
+    log = "log"
+
+    def __str__(self) -> str:
+        return self.value
+
+
+@attr.s(slots=True, repr=False)
+class StepRecord:
+    """Record within a step of a single test."""
+
+    record_id: str = attr.ib()
+    type: StepRecordType = attr.ib()
+    summary: Optional[str] = attr.ib()
+
+
+class StepLogRecord(LogRecord):
+    id: str = attr.ib(factory=(lambda: str(uuid4())))
+
+
+@attr.s(slots=True, repr=False)
+class Step:
+    """Step of a single test."""
+
+    title: str = attr.ib()
+    description: Optional[str] = attr.ib(factory=str)
+    status: Status = attr.ib(default=Status.success)
+    records: List[StepRecord] = attr.ib(factory=list)
+
+    def assertion(
+        self, assertion: AssertionResult, summary: Optional[str] = None
+    ) -> StepRecord:
+        record = StepRecord(
+            record_id=assertion.id,
+            type=StepRecordType.assertion,
+            summary=summary,
+        )
+        self.records.append(record)
+        self.status += assertion.status
+        return record
+
+    def interaction(
+        self, interaction: Interaction, summary: Optional[str] = None
+    ) -> StepRecord:
+        record = StepRecord(
+            record_id=interaction.id,
+            type=StepRecordType.interaction,
+            summary=summary,
+        )
+        self.records.append(record)
+        self.status += interaction.status
+        return record
+
+    def log(self, log: LogRecord, summary: Optional[str] = None) -> StepRecord:
+        record = StepRecord(
+            record_id=log.id,  # type: ignore
+            type=StepRecordType.log,
+            summary=summary,
+        )
+        self.records.append(record)
+        if log.levelno > ERROR:
+            self.status = Status.error
+        return record
+
+
 @attr.s(slots=True, repr=False)
 class TestResult:
     """Result of a single test."""
 
     __test__ = False
-    checks: List[Check] = attr.ib(factory=list)
+    assertions: Dict[str, AssertionResult] = attr.ib(factory=dict)
     errors: List[Exception] = attr.ib(factory=list)
-    interactions: List[Interaction] = attr.ib(factory=list)
-    logs: List[LogRecord] = attr.ib(factory=list)
+    interactions: Dict[str, Interaction] = attr.ib(factory=dict)
+    logs: Dict[str, LogRecord] = attr.ib(factory=dict)
+    steps: List[Step] = attr.ib(factory=list)
     is_errored: bool = attr.ib(default=False)
     summary: Optional[str] = attr.ib(default=None)
     # To show a proper reproduction code if an error happens and there is no way to get actual headers that were
@@ -168,70 +272,23 @@ class TestResult:
 
     def __add__(self, other: TestResult) -> TestResult:
         return TestResult(
-            checks=self.checks + other.checks,
+            assertions={**self.assertions, **other.assertions},
             errors=self.errors + other.errors,
-            interactions=self.interactions + other.interactions,
-            logs=self.logs + other.logs,
+            interactions={**self.interactions, **other.interactions},
+            logs={**self.logs, **other.logs},
             is_errored=self.is_errored or other.is_errored,
             summary=self.summary,
         )
 
-    def mark_errored(self) -> None:
-        self.is_errored = True
-
     @property
     def has_errors(self) -> bool:
         return bool(self.errors)
 
     @property
     def has_failures(self) -> bool:
-        return any(check.value == Status.failure for check in self.checks)
-
-    @property
-    def has_logs(self) -> bool:
-        return bool(self.logs)
-
-    def add_success(
-        self, name: str, response: requests.Response, elapsed: float
-    ) -> Check:
-        check = Check(
-            name=name,
-            value=Status.success,
-            response=response,
-            elapsed=elapsed,
-            request=None,
+        return any(
+            assertion.status == Status.failure for assertion in self.assertions.values()
         )
-        self.checks.append(check)
-        return check
-
-    def add_failure(
-        self,
-        name: str,
-        response: Optional[requests.Response],
-        elapsed: float,
-        message: str,
-        context: Optional[FailureContext],
-        request: Optional[requests.PreparedRequest] = None,
-    ) -> Check:
-        check = Check(
-            name=name,
-            value=Status.failure,
-            response=response,
-            elapsed=elapsed,
-            message=message,
-            context=context,
-            request=request,
-        )
-        self.checks.append(check)
-        return check
-
-    def add_error(self, exception: Exception) -> None:
-        self.errors.append(exception)
-
-    def store_requests_response(
-        self, response: requests.Response, status: Status, checks: List[Check]
-    ) -> None:
-        self.interactions.append(Interaction.from_requests(response, status, checks))
 
 
 @attr.s(slots=True)
@@ -253,32 +310,118 @@ class SerializedError:
         )
 
 
+@attr.s(slots=True)
+class SerializedAssertionResult:
+    name: str = attr.ib()
+    status: str = attr.ib()
+    interactions: List[SerializedInteraction] = attr.ib()
+    elapsed: float = attr.ib()
+    confidence: str = attr.ib()
+    risk: str = attr.ib()
+    recorded_at: str = attr.ib()
+    evidence: Optional[Evidence] = attr.ib()
+    reference: Optional[str] = attr.ib()
+    solution: Optional[str] = attr.ib()
+    message: Optional[str] = attr.ib()
+    code: Optional[str] = attr.ib()
+    cwe: Optional[str] = attr.ib()
+
+    @classmethod
+    def from_assertion_result(
+        cls, assertion_result: AssertionResult
+    ) -> "SerializedAssertionResult":
+        return cls(
+            name=assertion_result.name,
+            status=str(assertion_result.status),
+            interactions=[
+                SerializedInteraction.from_interaction(interaction)
+                for interaction in assertion_result.interactions
+            ],
+            elapsed=assertion_result.elapsed,
+            confidence=str(assertion_result.confidence),
+            risk=str(assertion_result.risk),
+            recorded_at=assertion_result.recorded_at,
+            evidence=assertion_result.evidence,
+            reference=assertion_result.reference,
+            solution=assertion_result.solution,
+            message=assertion_result.message,
+            code=assertion_result.code,
+            cwe=assertion_result.cwe,
+        )
+
+
 @attr.s(slots=True)
 class SerializedInteraction:
     request: Request = attr.ib()
     response: Response = attr.ib()
-    status: Status = attr.ib()
+    status: str = attr.ib()
+    elapsed: float = attr.ib()
     recorded_at: str = attr.ib()
+    curl_code: str = attr.ib()
+    name: Optional[str] = attr.ib()
 
     @classmethod
     def from_interaction(cls, interaction: Interaction) -> "SerializedInteraction":
         return cls(
             request=interaction.request,
             response=interaction.response,
-            status=interaction.status,
+            status=str(interaction.status),
             recorded_at=interaction.recorded_at,
+            elapsed=interaction.elapsed,
+            curl_code=interaction.request.as_curl_command(),
+            name=interaction.name,
+        )
+
+
+@attr.s(slots=True)
+class SerializedStepRecord:
+    record_id: str = attr.ib()
+    type: str = attr.ib()
+    summary: str = attr.ib()
+
+    @classmethod
+    def from_step_record(cls, step_record: StepRecord) -> "SerializedStepRecord":
+        return cls(
+            record_id=str(step_record.record_id),
+            type=str(step_record.type),
+            summary=step_record.summary if step_record.summary else "",
+        )
+
+
+@attr.s(slots=True)
+class SerializedStep:
+    title: str = attr.ib()
+    description: str = attr.ib()
+    records: List[SerializedStepRecord] = attr.ib(factory=list)
+    status: str = attr.ib(factory=(lambda: str(Status.success)))
+
+    @classmethod
+    def from_step(cls, step: Step) -> "SerializedStep":
+        return cls(
+            title=step.title,
+            description=step.description if step.description else "",
+            records=[
+                SerializedStepRecord.from_step_record(record) for record in step.records
+            ],
+            status=str(step.status),
         )
 
 
 @attr.s(slots=True)
 class SerializedTestResult:
+    assertions: Dict[str, SerializedAssertionResult] = attr.ib()
+    errors: List[SerializedError] = attr.ib()
+    interactions: Dict[str, SerializedInteraction] = attr.ib()
+    logs: Dict[str, dict] = attr.ib()
+    steps: List[SerializedStep] = attr.ib()
+    summary: Optional[str] = attr.ib()
+    # To show a proper reproduction code if an error happens and there is no way to get actual headers that were
+    # sent over the network. Or there could be no actual requests at all
+    overridden_headers: Optional[Dict[str, Any]] = attr.ib()
     has_failures: bool = attr.ib()
     has_errors: bool = attr.ib()
     has_logs: bool = attr.ib()
     is_errored: bool = attr.ib()
-    logs: List[str] = attr.ib()
-    errors: List[SerializedError] = attr.ib()
-    interactions: List[SerializedInteraction] = attr.ib()
 
     @classmethod
     def from_test_result(cls, result: TestResult) -> "SerializedTestResult":
@@ -288,14 +431,31 @@ class SerializedTestResult:
         return cls(
             has_failures=result.has_failures,
             has_errors=result.has_errors,
-            has_logs=result.has_logs,
+            has_logs=len(result.logs) > 0,
             is_errored=result.is_errored,
-            logs=[formatter.format(record) for record in result.logs],
+            logs={
+                str(key): {
+                    "message": logging.Formatter("%(message)s").format(record),
+                    "formatted": formatter.format(record),
+                    "level": record.levelname,
+                    "created": record.created,
+                }
+                for key, record in result.logs.items()
+            },
             errors=[SerializedError.from_error(error) for error in result.errors],
-            interactions=[
-                SerializedInteraction.from_interaction(interaction)
-                for interaction in result.interactions
-            ],
+            interactions={
+                str(interaction_id): SerializedInteraction.from_interaction(interaction)
+                for interaction_id, interaction in result.interactions.items()
+            },
+            assertions={
+                str(assertion_id): SerializedAssertionResult.from_assertion_result(
+                    assertion_result
+                )
+                for assertion_id, assertion_result in result.assertions.items()
+            },
+            steps=[SerializedStep.from_step(step) for step in result.steps],
+            summary=result.summary,
+            overridden_headers=result.overridden_headers,
         )
 
 
@@ -382,10 +542,9 @@ class AfterTestExecutionPayload(Payload):
 @attr.s(slots=True)
 class AfterStepExecutionPayload(Payload):
     test_case_id: str = attr.ib()
-    name: str = attr.ib()
     sequence: int = attr.ib()
     elapsed_time: float = attr.ib()
-    interaction: Interaction = attr.ib()
+    step: Step = attr.ib()
 
 
 @attr.s(slots=True)
